package YUV4MPEG;

use 5.006;
use strict;
use warnings;

require Exporter;
require DynaLoader;

# Attempt to load the PDL.  If we can't it's not the end of the world.

eval {
  require PDL::Core;
  import PDL::Core;
};

our @ISA = qw(Exporter DynaLoader);

our @EXPORT = qw(
  Y4M_OK Y4M_ERR_RANGE Y4M_ERR_SYSTEM Y4M_ERR_HEADER Y4M_ERR_BADTAG
  Y4M_ERR_MAGIC Y4M_ERR_EOF Y4M_ERR_XXTAGS Y4M_ERR_BADEOF Y4M_UNKNOWN	
);

our @EXPORT_OK = qw(
  Y4M_ILACE_NONE Y4M_ILACE_TOP_FIRST Y4M_ILACE_BOTTOM_FIRST
  FPS_UNKNOWN FPS_NTSC_FILM FPS_FILM FPS_PAL FPS_NTSC FPS_30 
  FPS_PAL_FIELD FPS_NTSC_FIELD FPS_60
  SAR_UNKNOWN SAR_SQUARE SAR_NTSC_CCIR601 SAR_NTSC_16_9 
  SAR_NTSC_SVCD_4_3 SAR_NTSC_SVCD_16_9 SAR_PAL_CCIR601 SAR_PAL_16_9
  SAR_PAL_SVCD_4_3 SAR_PAL_SVCD_16_9 SAR_SQR_ANA16_9
  DAR_UNKNOWN DAR_4_3 DAR_16_9 DAR_221_100
  SSM_UNKNOWN SSM_420_JPEG SSM_420_MPEG2
);

our %EXPORT_TAGS = ( 'all' => [ @EXPORT, @EXPORT_OK ] );

our $VERSION = '0.021';

sub Y4M_OK             {  0 }
sub Y4M_ERR_RANGE      {  1 }
sub Y4M_ERR_SYSTEM     {  2 }
sub Y4M_ERR_HEADER     {  3 }
sub Y4M_ERR_BADTAG     {  4 }
sub Y4M_ERR_MAGIC      {  5 }
sub Y4M_ERR_EOF        {  6 }
sub Y4M_ERR_XXTAGS     {  7 }
sub Y4M_ERR_BADEOF     {  8 }
sub Y4M_UNKNOWN        { -1 }

sub Y4M_ILACE_NONE         { 0 }
sub Y4M_ILACE_TOP_FIRST    { 1 }
sub Y4M_ILACE_BOTTOM_FIRST { 2 }

sub FPS_UNKNOWN        { _ratio_constant() }
sub FPS_NTSC_FILM      { _ratio_constant() }
sub FPS_FILM           { _ratio_constant() }
sub FPS_PAL            { _ratio_constant() }
sub FPS_NTSC           { _ratio_constant() }
sub FPS_30             { _ratio_constant() }
sub FPS_PAL_FIELD      { _ratio_constant() }
sub FPS_NTSC_FIELD     { _ratio_constant() }
sub FPS_60             { _ratio_constant() }

sub SAR_UNKNOWN        { _ratio_constant() }
sub SAR_SQUARE         { _ratio_constant() }
sub SAR_NTSC_CCIR601   { _ratio_constant() }
sub SAR_NTSC_16_9      { _ratio_constant() }
sub SAR_NTSC_SVCD_4_3  { _ratio_constant() }
sub SAR_NTSC_SVCD_16_9 { _ratio_constant() }
sub SAR_PAL_CCIR601    { _ratio_constant() }
sub SAR_PAL_16_9       { _ratio_constant() }
sub SAR_PAL_SVCD_4_3   { _ratio_constant() }
sub SAR_PAL_SVCD_16_9  { _ratio_constant() }
sub SAR_SQR_ANA16_9    { _ratio_constant() }

sub DAR_UNKNOWN        { _ratio_constant() }
sub DAR_4_3            { _ratio_constant() }
sub DAR_16_9           { _ratio_constant() }
sub DAR_221_100        { _ratio_constant() }

sub SSM_UNKNOWN        { 0 }
sub SSM_420_JPEG       { 1 }
sub SSM_420_MPEG2      { 2 }

sub _ratio_constant
{
    my $const_name = shift;
    my $symref;
    my ($package, $filename, $line, $sub) = caller(1);
    my @parts;

    $sub =~ s/^.*:://;

    @parts = split("_", $sub);
    $parts[0] = lc($parts[0]);    
    $sub = "y4m_" . join("_", @parts);

    $symref = DynaLoader::dl_find_symbol_anywhere("$sub");

    return YUV4MPEG::_load_ratio_constant($symref);
}

bootstrap YUV4MPEG $VERSION;

1;
__END__

=head1 NAME

YUV4MPEG - Perl API for reading/writing YUV4MPEG streams.

=head1 SYNOPSIS

  use YUV4MPEG;

  my $si = new YUV4MPEG::StreamInfo;
  my $fi = new YUV4MPEG::FrameInfo;
  my $frame;
  my $rgb;
 
  $si->read($filein); 
  $frame = new YUV4MPEG::Frame($si);

  $frame->read($fi, $filein);

  $frame->to_rgb();
  $rgb = $frame->get_packed_data();
  $frame->set_packed_data($rgb);
  $frame->to_yuv();

  $frame->write($fi, $fileout);

=head1 DESCRIPTION

YUV4MPEG is an interchange format created by the mjpegtools developers
for transporting video data in a pipeline between various applications
and tools.  This library provides an API for creating tools which can
read and write such streams, allowing one to creating video 
filters/processors in Perl.

The API is object oriented, and based around five main classes.  These
are:

  YUV4MPEG::StreamInfo
  YUV4MPEG::FrameInfo
  YUV4MPEG::Frame
  YUV4MPEG::Ratio
  YUV4MPEG::XTagList

The objects map roughly to the basic data structures in the yuv4mpeg
API, with the exception of the Frame class.  A basic program which
processes a stream from standard input and sends it to standard output
is as follows:

  use YUV4MPEG;

  my $si = new YUV4MPEG::StreamInfo;
  my $fi = new YUV4MPEG::FrameInfo;
  my $frame;
 
  $si->read($filein); 
  $frame = new YUV4MPEG::Frame($si);

  $frame->read($fi, $filein);
  $frame->write($fi, $fileout);

The StreamInfo class is used to store stream-level information.  This
includes frame rate, frame dimensions, and interlacing mode.  The
FrameInfo class, which is populated each time a frame is read, is
used to store frame data, such as tag information.  And the Frame
class itself provides an abstraction of the low-level frame data, and
provides support functions for retrieving and manipulating that data.

So, these three classes, StreamInfo, FrameInfo, and Frame, make up the
primary objects of the YUV4MPEG API.  In addition, there are the Ratio
and XTagList classes.  These are containers for ratio data (frame
rates, aspect ratios, etc) and tagging information, respectively.

=head1 CLASS DESCRIPTIONS

The following are descriptions of each of the various classes 
available.

=head2 YUV4MPEG::StreamInfo

The StreamInfo classes provides a container for all stream-level
information.

=over 8

=item new ( [ SI ] )

Creates a new StreamInfo class, which may be a copy of the provided
StreamInfo object, if one is given.

=item set_width ( WIDTH )

Sets the width of the video stream to WIDTH.

=item get_width

Retrieves the width of the video stream.

=item set_height ( HEIGHT )

Sets the height of the video stream to HEIGHT.

=item get_height

Retrieves the height of the video stream.

=item set_interlace ( INTERLACE )

Sets the interlacing mode.  Assuming constants are imported, the
following are valid:

=over 8

=item Y4M_ILACE_NONE

No interlacing/progressive mode.

=item Y4M_ILACE_TOP_FIRST

Top-field first interlacing.

=item Y4M_ILACE_BOTTOM_FIRST

Bottom-field first interlacing.

=back

=item get_interlace

Gets the interlacing mode for this stream.

=item set_framerate ( RATIO )

Sets the frame rate for this stream.  RATIO must be an instance of
the YUV4MPEG::Ratio class.

=item get_framerate

Gets the frame rate for this stream.

=item set_sampleaspect ( RATIO )

Sets the aspect ratio for this stream.

=item get_sampleaspect

Gets the aspect ratio for this stream.

=item get_framelength

Get the number of bytes per frame of data (excluding headers) for this
stream.

=item xtags

Get the xtags for this stream.  This returns a YUV4MPEG::XTagList object.

=item parse ( TAGS )

Parse a set of tags from a string.

=item read ( [ HANDLE ] )

Read stream information from the specified HANDLE.  If no handle is
given, STDIN is assumed.

=item write ( [ HANDLE ] )

Write a stream header out to the specified HANDLE.  If no handle is
given, STDOUT is assumed.

=back

=head2 YUV4MPEG::FrameInfo

The FrameInfo structure is used for storing frame-level data.  Currently,
this is only xtags.

=over 8

=item new ( [ FI ] )

Create a new YUV4MPEG::FrameInfo instance.  If FI is defined, and is
an instance of YUV4MPEG::FrameInfo, then the new instance is made to
be a copied of the provided instance.

=item xtags

Get the xtags for this FrameInfo instance.  Note, depending on your
version of mjpegtools, this method may not be available.  In version
1.6.1, and all previous version, this method was left unimplemented
(an oversight on the developers part).  However, this has been
remedied in later versions.

=item read ( [ HANDLE ] )

Read a frame header into this FrameInfo instance from the specified
HANDLE, or STDIN if none is given.

=item write ( [ HANDLE ] )

Write out a frame header to the specified HANDLE, or STDOUT if none
is given.

=back

=head2 YUV4MPEG::Frame

The Frame class wraps the low-level frame data read from a stream, and
provides useful interfaces to set/get this data in various formats and
perform certain operations on that data (eg, colorspace conversions).

Note that this class also does the work of interlacing/deinterlacing
frame data on the fly.  If frame data is read, and it is interlaced,
then the data is deinterlaced upon the first get_* call.  When the
frame is about to be written out, if the data is deinterlaced (either
by a get_ call, or because new data has been written with a set_
call), it is interlaced as necessary before it is written.  Thus, the
user need not worry about the interlacing mode of the data.

Now, currently there are no methods for getting only the top or
bottom fields of a given frame.  However, if that is desireable, 
access methods can be written to facilitate this.

Anyway, on to the methods.

=over 8

=item new ( STREAM )

Create a new YUV4MPEG::Frame instance.  STREAM must be an instance of
YUV4MPEG::StreamInfo, and is used to get the frame dimensions and
stream interlacing mode.

=item read ( FI, [ HANDLE ] )

Read a single frame from from the HANDLE specified, or STDIN if none
is given.  This method populates the FrameInfo instance with
information about the frame read (eg, tags).

=item write ( FI, [ HANDLE ] )

Write a single frame to the HANDLE specified, or STDOUT if none is
given.  The information in the FrameInfo structure is outputted in the
header of the frame data emitted.

=item get_data

Returns three array references, which represent the three data planes
in the frame.

=item set_data ( P1, P2, P3 )

Stores the frame data provided into the Frame object.  In this case,
P1, P2, and P3 must be array references which contain the three planes
in the image.

=item get_combined_data

Retrieves the frame data as a single array, with the data planes
combined, and returns a reference to this array.  In this case, the
data is interleaved together.  eg, if P1, P2, and P3 are the planes,
the array returned will be:

( P1[0], P2[0], P3[0], P1[1], P2[1], P3[1], ... )

=item set_combined_data ( DATA )

Stores the provided data in the frame object.  DATA must be an array
reference, with the data interleaved as described above.

=item get_packed_data

Retrieves the frame data as a single string.  This string is the same
as that generated by:

$data = pack("C*", @{ $frame->get_combined_data() })

=item set_packed_data ( DATA )

Stores the provided data in the frame object, where DATA is a string
describing the frame, formatted as described above.

=item get_pdl_data

If the PDL is installed, you can get the frame data as a regular ol'
piddle, using this method.  This returns a 3 x Width x Height piddle
of bytes.

=item set_pdl_data ( [ PIDDLE ] )

Sets the data in the frame to that stored in the provided piddle.
The piddle must be 3 x Width x Height or an error will be thrown.

=item to_rgb ( [ MODE ] )

Converts the frame data from YUV to RGB.

This function first can perform supersampling of the chroma data in
the stream, so it is not necessary to do this by hand first.  The
supersampling mode is specified with the optional MODE parameter, and
can be one of SSM_UNKNOWN, SSM_420_JPEG, or SSM_420_MJPEG2, with the
default being SSM_420_JPEG.  Specifying SSM_UNKNOWN disables
supersampling.

=item to_yuv ( [ MODE ] )

Converts the frame data from RGB to YUV.

This function can perform subsampling of the chroma data in the
stream, so it is not necessary to do this by hand after the
conversion.  The subsampling mode is specified with the optional
MODE parameter, and can be one of SSM_UNKNOWN, SSM_420_JPEG, or
SSM_420_MJPEG2, with the default being SSM_420_JPEG.  Specifying
SSM_UNKNOWN disables subsampling.

=item chroma_supersample ( [ MODE ] )

Performs chroma supersampling on the YUV data in the frame object.

=item chroma_subsample ( [ MODE ] )

Performs chroma subsampling on the YUV data in the frame object.

=back

=head2 YUV4MPEG::Ratio

This class is used to deal with various ratios.  These can include
frame rates, aspect ratios, and so forth.  Each ratio has an 
integer numerator and denominator, and there are a few miscellaneous
operations which can be performed on them.

=over 8

=item new ( [ N, D ] )

Create a new YUV4MPEG::Ratio instance.  The numerator and denominator
are optional, and default to 0 and 1 if left unspecified.

=item get_n

Get the numerator portion of the ratio.

=item get_d

Get the denominator portion of the ratio.

=item get

Returns an array containing the numerator and denominator.

=item set_n ( N )

Set the numerator of the ratio.

=item set_d ( D )

Set the denominator of the ratio.  Zero is disallowed.

=item set ( N, D )

Set the numerator and denominator of the ratio.  Zero is disallowed as
a denominator.

=item reduce

Reduces the ratio.

=item parse ( SCALAR )

Parses the string specified, which must be of the form "n:d" and sets
the numerator and denominator accordingly.

=item equals ( RATIO )

Determines if the YUV4MPEG::Ratio instance specified is equal to this one.
Returns 1 if it is, and 0 if not.  This is based strictly on a comparison
of the numerator and denominator (no reduction is done first).

=item guess_sar ( RATIO, W, H )

Given the display aspect ratio provided, and the width and height, attempts
to guess the sample aspect ratio of the stream, based on a set of known
values.

=back

=head2 YUV4MPEG::XTagList

The XTagList, which is only obtainable from StreamInfo or FrameInfo objects,
is simply a list of XTags, or user-definable metadata, which is attached to
an object.  It is essentially a glorified array of strings.

=over 8

=item copy ( SRC )

Makes the current list a copy of the provided one.

=item count

Retrieves the number of entries in the XTagList.

=item get ( N )

Retrieves the n'th entry in the XTagList.

=item add ( STR )

Adds the provided string to the end of the XTagList.

=item remove ( N )

Removes the n'th entry from the XTagList.

=item clearlist

Clears the list of all entries.

=item addlist ( SRC )

Adds the tags from the provided XTagList to the end of this one.

=back

=head2 MISCELLANEOUS FUNCTIONS

There are a number of functions which exist in the YUV4MPEG package
which do not really fall under any specific catagory.  Hence, they are
described here.

=over 8

=item YUV4MPEG::y4m_strerr ( N )

Converts the provided YUV4MPEG error number into a semi-readable string.

=item YUV4MPEG::y4m_allow_unknown_tags ( BOOL )

Sets whether or not we allow unknown tags when we're reading/writing
YUV4MPEG streams.  Returns the current setting.

=back

=head1 AUTHOR

Brett Kosinski, E<lt>brettk@frodo.dyn.gno.orgE<gt>

=cut
